# Imports from other dependencies.
from tqdm._utils import _term_move_up


# Imports from geography.
from geography.models import Division
from geography.models import DivisionLevel


class Attributes(object):
    def set_attributes(self):
        self.NATIONAL_LEVEL, created = DivisionLevel.objects.get_or_create(
            name=DivisionLevel.COUNTRY
        )
        self.STATE_LEVEL, created = DivisionLevel.objects.get_or_create(
            name=DivisionLevel.STATE, parent=self.NATIONAL_LEVEL
        )
        self.DISTRICT_LEVEL, created = DivisionLevel.objects.get_or_create(
            name=DivisionLevel.DISTRICT, parent=self.STATE_LEVEL
        )
        self.COUNTY_LEVEL, created = DivisionLevel.objects.get_or_create(
            name=DivisionLevel.COUNTY, parent=self.STATE_LEVEL
        )

        # Other fixtures
        self.TOWNSHIP_LEVEL, created = DivisionLevel.objects.get_or_create(
            name=DivisionLevel.TOWNSHIP, parent=self.COUNTY_LEVEL
        )
        self.PRECINCT_LEVEL, created = DivisionLevel.objects.get_or_create(
            name=DivisionLevel.PRECINCT, parent=self.COUNTY_LEVEL
        )
        self.VOTERS_ABROAD_LEVEL, created = DivisionLevel.objects.get_or_create(
            name=DivisionLevel.VOTERS_ABROAD
        )

        self.NATION, created = Division.objects.get_or_create(
            code="00",
            name="United States of America",
            label="United States of America",
            short_label="USA",
            level=self.NATIONAL_LEVEL,
        )

        self.TQDM_PREFIX = _term_move_up() + "\r"
        self.SHP_SOURCE_BASE = "https://www2.census.gov/geo/tiger/GENZ{}/shp/"
        self.DOWNLOAD_DIRECTORY = "./tmp/data/geography/"



# Imports from Django.
from django.core.management.base import BaseCommand


# Imports from other dependencies.
from tqdm import tqdm


# Imports from geography.
from geography.management.commands.bake._arguments import ArgumentsMethods
from geography.management.commands.bake._attributes import Attributes
from geography.management.commands.bake.geometries import DistrictGeometries
from geography.management.commands.bake.geometries import StateGeometries


class Command(
    ArgumentsMethods,
    Attributes,
    StateGeometries,
    DistrictGeometries,
    BaseCommand,
):
    help = (
        "Uploads topojson files by state and district to an Amazon S3 bucket."
    )

    def handle(self, *args, **options):
        self.parse_arguments(options)
        self.set_attributes()

        tqdm.write("🌎  BAKING GEOGRAPHY 🌎\n")

        self.bake_state_geometries()
        self.bake_district_geometries()

        self.stdout.write(self.style.SUCCESS("All done! 🏁"))



# Imports from geography.
from geography.management.commands.bootstrap.fixtures._county import (
    CountyFixtures
)
from geography.management.commands.bootstrap.fixtures._district import (
    DistrictFixtures
)
from geography.management.commands.bootstrap.fixtures._nation import (
    NationFixtures
)
from geography.management.commands.bootstrap.fixtures._state import (
    StateFixtures
)



# Imports from Django.
from django.contrib.postgres.fields import JSONField
from django.core.exceptions import ObjectDoesNotExist
from django.db import models


# Imports from other dependencies.
from civic_utils.models import CivicBaseModel
from civic_utils.models import CommonIdentifiersMixin
from civic_utils.models import UUIDMixin


# Imports from geography.
from geography.models.division_level import DivisionLevel
from geography.models.intersect_relationship import IntersectRelationship


class Division(CommonIdentifiersMixin, UUIDMixin, CivicBaseModel):
    """A political or administrative geography.

    For example, a certain state, county, district, precinct or municipality.
    """

    natural_key_fields = ["level", "uid"]
    uid_prefix = "division"
    uid_base_field = "name"
    # default_serializer = ""

    name = models.CharField(max_length=255)

    label = models.CharField(max_length=255, blank=True)
    short_label = models.CharField(max_length=50, null=True, blank=True)

    parent = models.ForeignKey(
        "self",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="children",
    )

    level = models.ForeignKey(
        DivisionLevel, on_delete=models.PROTECT, related_name="divisions"
    )

    code = models.CharField(
        max_length=200,
        help_text="Code representing a geography: FIPS code for states and \
        counties, district number for districts, precinct number for \
        precincts, etc.",
    )
    code_components = JSONField(
        blank=True, null=True, help_text="Component parts of code"
    )

    intersecting = models.ManyToManyField(
        "self",
        through="IntersectRelationship",
        symmetrical=False,
        related_name="+",
        help_text="Intersecting divisions intersect this one geographically "
        "but do not necessarily have a parent/child relationship. "
        "The relationship between a congressional district and a "
        "precinct is an example of an intersecting relationship.",
    )

    effective = models.BooleanField(default=True)
    effective_start = models.DateTimeField(null=True, blank=True)
    effective_end = models.DateTimeField(null=True, blank=True)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        """
        **uid**: :code:`division:{parentuid}_{levelcode}-{code}`
        """
        self.generate_common_identifiers()

        super(Division, self).save(*args, **kwargs)

    def get_uid_suffix(self):
        current_division_slug = "{}={}".format(self.level.slug, self.code)

        if self.parent:
            return "{}&{}".format(
                self.parent.get_uid_suffix(), current_division_slug
            )

        return current_division_slug

    def add_intersecting(self, division, intersection=None, symm=True):
        """
        Adds paired relationships between intersecting divisions.

        Optional intersection represents the portion of the area of the related
        division intersecting this division. You can only specify an
        intersection on one side of the relationship when adding a peer.
        """
        relationship, created = IntersectRelationship.objects.update_or_create(
            from_division=self,
            to_division=division,
            defaults={"intersection": intersection},
        )
        if symm:
            division.add_intersecting(self, None, False)
        return relationship

    def remove_intersecting(self, division, symm=True):
        """Removes paired relationships between intersecting divisions"""
        IntersectRelationship.objects.filter(
            from_division=self, to_division=division
        ).delete()
        if symm:
            division.remove_intersecting(self, False)

    def set_intersection(self, division, intersection):
        """Set intersection percentage of intersecting divisions."""
        IntersectRelationship.objects.filter(
            from_division=self, to_division=division
        ).update(intersection=intersection)

    def get_intersection(self, division):
        """Get intersection percentage of intersecting divisions."""
        try:
            return IntersectRelationship.objects.get(
                from_division=self, to_division=division
            ).intersection
        except ObjectDoesNotExist:
            raise Exception("No intersecting relationship with that division.")

    def has_elections(self, date):
        if len(self.elections.filter(election_day__date=date)) > 0:
            return True

        for child in self.children.all():
            if len(child.elections.filter(election_day__date=date)) > 0:
                return True

        return False



# Imports from python.
import json
import os


# Imports from Django.
from django.contrib.humanize.templatetags.humanize import ordinal


# Imports from other dependencies.
from tqdm import tqdm
import shapefile
from uuslug import slugify

# Imports from geography.
from geography.models import Division
from geography.models import Geometry


class DistrictFixtures(object):
    def create_district_fixtures(self):
        SHP_SLUG = "cb_{}_us_cd{}_500k".format(self.YEAR, self.CONGRESS)
        DOWNLOAD_PATH = os.path.join(self.DOWNLOAD_DIRECTORY, SHP_SLUG)

        shape = shapefile.Reader(
            os.path.join(DOWNLOAD_PATH, "{}.shp".format(SHP_SLUG))
        )
        fields = shape.fields[1:]
        field_names = [f[0] for f in fields]

        for shp in tqdm(shape.shapeRecords(), desc="Districts"):
            district = dict(zip(field_names, shp.record))
            # Skip territories and DC
            if int(district["STATEFP"]) > 56 or int(district["STATEFP"]) == 11:
                continue

            state = Division.objects.get(
                code=district["STATEFP"], level=self.STATE_LEVEL
            )
            code_key = "CD{}FP".format(self.CONGRESS)
            if int(district[code_key]) == 0:
                label = "{} at-large congressional district".format(
                    state.label, ordinal(int(district[code_key]))
                )
            else:
                label = "{} {} congressional district".format(
                    state.label, ordinal(int(district[code_key]))
                )
            district_obj, created = Division.objects.update_or_create(
                code=district[code_key],
                level=self.DISTRICT_LEVEL,
                parent=state,
                defaults={
                    "name": label,
                    "label": label,
                    "code_components": {
                        "fips": {"state": state.code},
                        "district": district[code_key],
                    },
                },
            )
            geodata = {
                "type": "Feature",
                "geometry": shp.shape.__geo_interface__,
                "properties": {
                    "state": state.code,
                    "district": district[code_key],
                    "name": label,
                },
            }
            geojson, created = Geometry.objects.update_or_create(
                division=district_obj,
                subdivision_level=self.DISTRICT_LEVEL,
                simplification=self.THRESHOLDS["district"],
                data_summary=slugify(
                    "{}--{}--{}".format(
                        district_obj.slug,
                        self.DISTRICT_LEVEL.slug,
                        self.THRESHOLDS["district"],
                    )
                ),
                source=os.path.join(
                    self.SHP_SOURCE_BASE.format(self.YEAR), SHP_SLUG
                )
                + ".zip",
                series=self.YEAR,
                defaults={
                    "topojson": self.toposimplify(
                        geodata, self.THRESHOLDS["district"]
                    )
                },
            )
            tqdm.write(
                self.TQDM_PREFIX
                + ">  FIPS {}, District {}  @ ~{}kb               ".format(
                    district["STATEFP"],
                    district[code_key],
                    round(len(json.dumps(geojson.topojson)) / 1000),
                )
            )

            geojson, created = Geometry.objects.update_or_create(
                division=district_obj,
                subdivision_level=self.COUNTY_LEVEL,
                simplification=self.THRESHOLDS["county"],
                data_summary=slugify(
                    "{}--{}--{}".format(
                        district_obj.slug,
                        self.COUNTY_LEVEL.slug,
                        self.THRESHOLDS["county"],
                    )
                ),
                source=os.path.join(
                    self.SHP_SOURCE_BASE.format(self.YEAR), SHP_SLUG
                )
                + ".zip",
                series=self.YEAR,
                defaults={
                    "topojson": self.get_district_county_shps(
                        district["STATEFP"], district[code_key]
                    )
                },
            )

            tqdm.write(
                self.TQDM_PREFIX
                + ">  FIPS {}, District {}  @ ~{}kb         ".format(
                    district["STATEFP"],
                    district[code_key],
                    round(len(json.dumps(geojson.topojson)) / 1000),
                )
            )
        tqdm.write(self.style.SUCCESS("Done.\n"))



# Imports from other dependencies.
from civic_utils.serializers import CommandLineListSerializer
from civic_utils.serializers import NaturalKeySerializerMixin
from rest_framework import serializers


# Imports from geography.
from geography.models import Geometry


class GeometryListSerializer(
    NaturalKeySerializerMixin, CommandLineListSerializer
):
    subdivision_level = serializers.SerializerMethodField()
    division = serializers.StringRelatedField()

    def get_subdivision_level(self, obj):
        return obj.subdivision_level.slug

    class Meta(CommandLineListSerializer.Meta):
        model = Geometry
        fields = ("id", "division", "subdivision_level")



# Imports from other dependencies.
from tqdm import tqdm


# Imports from geography.
from geography.models import Division
from geography.utils.lookups import counties


class CountyFixtures(object):
    def create_county_fixtures(self):
        for county in tqdm(counties, desc="Counties"):
            # Skip territories and DC
            if int(county["state"]) > 56 or int(county["state"]) == 11:
                continue
            state = Division.objects.get(
                code=county["state"], level=self.STATE_LEVEL
            )
            Division.objects.update_or_create(
                level=self.COUNTY_LEVEL,
                code="{}{}".format(county["state"], county["county"]),
                parent=state,
                defaults={
                    "name": county["NAME"],
                    "label": county["NAME"],
                    "code_components": {
                        "fips": {
                            "state": county["state"],
                            "county": county["county"],
                        }
                    },
                },
            )
            tqdm.write(
                self.TQDM_PREFIX
                + ">  FIPS {}{}     ".format(county["state"], county["county"])
            )
        tqdm.write(self.style.SUCCESS("Done.\n"))



# Imports from python.
import os


# Imports from other dependencies.
import geojson
import shapefile


# Imports from geography.
from geography.utils.lookups import township_lookup


class StateTownshipShapes(object):
    def get_state_township_shps(self, fips):
        SHP_SLUG = "cb_{}_{}_cousub_500k".format(self.YEAR, fips)
        DOWNLOAD_PATH = os.path.join(self.DOWNLOAD_DIRECTORY, SHP_SLUG)
        shape = shapefile.Reader(
            os.path.join(DOWNLOAD_PATH, "{}.shp".format(SHP_SLUG))
        )
        fields = shape.fields[1:]
        field_names = [f[0] for f in fields]
        township_records = [
            shp
            for shp in shape.shapeRecords()
            if dict(zip(field_names, shp.record))["STATEFP"] == fips
            or (
                fips == "00"
                and int(dict(zip(field_names, shp.record))["STATEFP"]) <= 56
            )
        ]
        features = []
        for shp in township_records:
            rec = dict(zip(field_names, shp.record))
            if rec["COUSUBFP"] == "00000":
                continue

            geometry = shp.shape.__geo_interface__
            geodata = {
                "type": "Feature",
                "geometry": geometry,
                "properties": {
                    "state": rec["STATEFP"],
                    "county": rec["COUNTYFP"],
                    "countysub": rec["COUSUBFP"],
                    "name": township_lookup[rec["STATEFP"]][
                        rec["COUNTYFP"]
                    ].get(rec["COUSUBFP"], rec["NAME"]),
                },
            }
            features.append(geodata)
        threshold = self.THRESHOLDS["county"]
        return self.toposimplify(
            geojson.FeatureCollection(features), threshold
        )



# Imports from Django.
from django.db import models


# Imports from other dependencies.
from civic_utils.models import CivicBaseModel
from civic_utils.models import UniqueIdentifierMixin
from civic_utils.models import UUIDMixin
from uuslug import slugify


class DivisionLevel(UniqueIdentifierMixin, UUIDMixin, CivicBaseModel):
    """Level of government or administration where a division exists.

    Exempli gratia, federal, state, district, county, precinct, municipal.
    """

    natural_key_fields = ["name"]
    uid_prefix = "divisionlevel"
    uid_base_field = "slug"
    # default_serializer = ""

    COUNTRY = "country"
    STATE = "state"
    DISTRICT = "district"
    COUNTY = "county"
    TOWNSHIP = "township"
    PRECINCT = "precinct"

    VOTERS_ABROAD = "abroad"

    LEVEL_CHOICES = (
        (COUNTRY, "Country"),
        (STATE, "State"),
        (DISTRICT, "District"),
        (COUNTY, "County"),
        (TOWNSHIP, "Township"),
        (PRECINCT, "Precinct"),
        (VOTERS_ABROAD, "Voters abroad"),
    )

    name = models.CharField(max_length=255, unique=True, choices=LEVEL_CHOICES)

    slug = models.SlugField(
        blank=True, max_length=255, unique=True, editable=False
    )

    parent = models.ForeignKey(
        "self",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="children",
    )

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name)

        self.generate_unique_identifier()

        super(DivisionLevel, self).save(*args, **kwargs)

    def get_uid_prefix(self):
        if self.parent:
            return "{}:{}".format(self.parent.uid, self.uid)

        return self.uid



# Generated by Django 2.0.1 on 2018-03-14 20:48

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('geography', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Point',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('lat', models.FloatField(help_text='Latitude coordinate in decimal degrees.')),
                ('lon', models.FloatField(help_text='Longitude coordinate in decimal degrees.')),
                ('attributes', django.contrib.postgres.fields.jsonb.JSONField(blank=True, help_text='Miscellaneous attributes on the point.', null=True)),
                ('threshold', models.PositiveSmallIntegerField(default=0, help_text='A threshold in pixels above which to display this point.')),
                ('label', models.CharField(max_length=250)),
                ('geometry', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='points', to='geography.Geometry')),
            ],
        ),
        migrations.CreateModel(
            name='PointLabelOffset',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('x', models.SmallIntegerField(default=0, help_text='Lateral offset in pixels.')),
                ('y', models.SmallIntegerField(default=0, help_text='Vertical offset in pixels.')),
                ('threshold', models.PositiveSmallIntegerField(default=0, help_text='A threshold in pixels above which to apply this offset.')),
                ('point', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='offsets', to='geography.Point')),
            ],
        ),
    ]



# Imports from other dependencies.
from rest_framework import viewsets


# Imports from geography.
from geography.conf import settings
from geography.utils.importers import import_class


authentication = import_class(settings.API_AUTHENTICATION_CLASS)
permission = import_class(settings.API_PERMISSION_CLASS)
pagination = import_class(settings.API_PAGINATION_CLASS)


class BaseViewSet(viewsets.ModelViewSet):
    authentication_classes = (authentication,)
    permission_classes = (permission,)
    pagination_class = pagination



# Imports from Django.
from django.core.management.base import BaseCommand


# Imports from other dependencies.
from tqdm import tqdm


# Imports from geography.
from geography.management.commands.bootstrap._arguments import ArgumentsMethods
from geography.management.commands.bootstrap._attributes import Attributes
from geography.management.commands.bootstrap._toposimplify import Toposimplify
from geography.management.commands.bootstrap.aggregate_shapes.district import (
    DistrictCountyShapes,
)
from geography.management.commands.bootstrap.aggregate_shapes.state import (
    StateCountyShapes,
    StateDistrictShapes,
    StateTownshipShapes,
)
from geography.management.commands.bootstrap.download import DownloadCounty
from geography.management.commands.bootstrap.download import DownloadDistrict
from geography.management.commands.bootstrap.download import DownloadState
from geography.management.commands.bootstrap.download import DownloadTownship
from geography.management.commands.bootstrap.fixtures import CountyFixtures
from geography.management.commands.bootstrap.fixtures import DistrictFixtures
from geography.management.commands.bootstrap.fixtures import NationFixtures
from geography.management.commands.bootstrap.fixtures import StateFixtures


class Command(
    ArgumentsMethods,
    Attributes,
    Toposimplify,
    DownloadState,
    DownloadCounty,
    DownloadDistrict,
    DownloadTownship,
    CountyFixtures,
    DistrictFixtures,
    NationFixtures,
    StateFixtures,
    StateCountyShapes,
    StateDistrictShapes,
    StateTownshipShapes,
    DistrictCountyShapes,
    BaseCommand,
):
    help = (
        "Downloads and bootstraps geographic data for states, congressional "
        "districts, counties and townships from the U.S. Census Bureau "
        "simplified cartographic boundary files."
    )

    def handle(self, *args, **options):
        self.set_attributes()
        self.parse_arguments(options)

        tqdm.write("🌎  BOOTSTRAPING GEOGRAPHY 🌎")
        tqdm.write("(Get a ☕ . This will take a few minutes.)\n")
        tqdm.write("Downloading data 📡")
        self.download_state_data()
        self.download_district_data()
        self.download_county_data()
        self.download_township_data()
        self.stdout.write(self.style.SUCCESS("Done.\n"))

        tqdm.write("Creating fixtures 🌐")
        self.create_nation_fixtures()
        self.create_state_fixtures()
        self.create_district_fixtures()
        self.create_county_fixtures()

        self.stdout.write(self.style.SUCCESS("All done! 🏁"))



# Imports from geography.
from geography.models.division_level import DivisionLevel
from geography.models.division import Division
from geography.models.geometry import Geometry
from geography.models.intersect_relationship import IntersectRelationship
from geography.models.point import Point
from geography.models.point import PointLabelOffset


__all__ = [
    "DivisionLevel",
    "Division",
    "Geometry",
    "IntersectRelationship",
    # "Point",
    # "PointLabelOffset",
]



# Imports from geography.
from geography.models import PointLabelOffset
from geography.serializers import PointLabelOffsetSerializer
from geography.viewsets.base import BaseViewSet


class PointLabelOffsetViewSet(BaseViewSet):
    queryset = PointLabelOffset.objects.all()
    serializer_class = PointLabelOffsetSerializer



# Imports from python.
import json
import uuid


# Imports from Django.
from django.contrib.postgres.fields import JSONField
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.safestring import mark_safe


# Imports from other dependencies.
from civic_utils.models import CivicBaseModel
from civic_utils.models import UniqueIdentifierMixin
from civic_utils.models import UUIDMixin


# Imports from geography.
from geography.models.division import Division
from geography.models.division_level import DivisionLevel


class Geometry(UniqueIdentifierMixin, UUIDMixin, CivicBaseModel):
    """The spatial representation (in TopoJSON) of a Division."""

    D3 = """
        <div id="map{0}"></div>
        <script>
        var data{0} = {1};
        var feature{0} = topojson.feature(
            data{0}, data{0}.objects.divisions);
        var projection{0} = d3.geoMercator().scale(1)
            .fitSize([{2}, {2}], feature{0});
        var points{0} = data{0}.objects.points ? topojson.feature(
            data{0}, data{0}.objects.points).features : [];
        var svg{0} = d3.select("#map{0}").append("svg")
            .attr("width", {2})
            .attr("height", {2});
        svg{0}.append("path").datum(feature{0})
            .attr("d", d3.geoPath().projection(
                projection{0}
            ));
        svg{0}.selectAll('circle.c{0}')
            .data(points{0})
          .enter().append('circle')
            .attr('class', 'c{0}')
            .attr('cx', function(d) {{
                return projection{0}(d.geometry.coordinates)[0];
            }})
            .attr('cy', function(d) {{
                return projection{0}(d.geometry.coordinates)[1];
            }})
            .attr('r', {2} < 100 ? 2 : 4)
            .attr('fill', 'transparent')
            .attr('stroke-width', {2} < 100 ? 1 : 2)
            .attr('stroke', 'black');
        svg{0}.selectAll('text.t{0}')
            .data(points{0})
        .enter().append('text')
            .text(function(d) {{
                return d.properties.label
            }})
            .attr('x', function(d) {{
                var offset = -10;
                if (d.properties.attributes) {{
                    offset = d.properties.attributes['0'][0];
                }}

                return projection{0}(d.geometry.coordinates)[0] + offset;
            }})
            .attr('y', function(d) {{
                var offset = 4;
                if (d.properties.attributes) {{
                    offset = d.properties.attributes['0'][1];
                }}

                return projection{0}(d.geometry.coordinates)[1] + offset;
            }})
            .attr('fill', 'black')
            .attr('text-anchor', 'end');
        </script>
    """
    natural_key_fields = ["division", "subdivision_level", "data_summary"]
    uid_prefix = "geometry"
    # default_serializer = ""

    division = models.ForeignKey(
        Division, on_delete=models.CASCADE, related_name="geometries"
    )

    subdivision_level = models.ForeignKey(
        DivisionLevel, on_delete=models.PROTECT, related_name="+"
    )

    data_summary = models.SlugField(
        blank=True,
        max_length=255,
        unique=True,
        editable=False,
        help_text=(
            "A brief description (written as a slug) of what makes this data "
            "unique from other Geometry instances with the same 'division' "
            "and 'subdivision_level' values."
        ),
    )

    simplification = models.FloatField(
        validators=[MinValueValidator(0.0), MaxValueValidator(1.0)],
        help_text=(
            "Minimum quantile of planar triangle areas for "
            "simplfying topojson."
        ),
    )

    topojson = JSONField()

    source = models.URLField(
        blank=True,
        null=True,
        help_text="Link to the source of this geography data.",
    )

    series = models.CharField(
        blank=True,
        null=True,
        max_length=4,
        help_text="Year of boundary series, e.g., 2016 TIGER/Line files.",
    )

    effective = models.BooleanField(default=True)

    effective_start = models.DateField(null=True, blank=True)

    effective_end = models.DateField(null=True, blank=True)

    class Meta:
        unique_together = ("division", "subdivision_level", "data_summary")
        verbose_name_plural = "Geometries"

    def __str__(self):
        return "{} - {} map, {}".format(
            self.division.label,
            self.subdivision_level.name,
            self.simplification,
        )

    def save(self, *args, **kwargs):
        self.generate_unique_identifier()

        super(Geometry, self).save(*args, **kwargs)

    def get_uid_suffix(self):
        return (
            f"division={self.division.uid}"
            f"&subdivision_level={self.subdivision_level.uid}"
            f"&data_summary={self.data_summary}"
        )

    def small_preview(self):
        return mark_safe(
            self.D3.format(
                self.pk.hex.replace("-", ""), self.to_topojson(), "60"
            )
        )

    def large_preview(self):
        return mark_safe(
            self.D3.format(
                self.pk.hex.replace("-", ""), self.to_topojson(), "400"
            )
        )

    def file_size(self):
        return "~{} kB".format(round(len(self.to_topojson()) / 1000))

    def to_topojson(self):
        """Adds points and converts to topojson string."""
        topojson = self.topojson
        topojson["objects"]["points"] = {
            "type": "GeometryCollection",
            "geometries": [point.to_topojson() for point in self.points.all()],
        }
        return json.dumps(topojson)



# Imports from other dependencies.
from civic_utils.serializers import CommandLineListSerializer
from civic_utils.serializers import NaturalKeySerializerMixin


# Imports from geography.
from geography.models import Point


class PointSerializer(NaturalKeySerializerMixin, CommandLineListSerializer):
    class Meta(CommandLineListSerializer.Meta):
        model = Point
        fields = "__all__"



# Imports from python.
import json
import os
import subprocess


# Imports from Django.
from django.contrib.humanize.templatetags.humanize import ordinal
from django.core.management.base import BaseCommand


# Imports from other dependencies.
import geojson
import shapefile
from tqdm import tqdm


# Imports from geography.
from geography.models import Division
from geography.models import DivisionLevel


SHP_BASE = "http://www.pacourts.us/assets/files/setting-6061/file-6845.zip"
SHP_SLUG = "clipped_pa"
STATE_LEVEL = DivisionLevel.STATE


class Command(BaseCommand):
    @staticmethod
    def toposimplify(geojson, p):
        """
        Convert geojson and simplify topology.

        geojson is a dict representing geojson.
        p is a simplification threshold value between 0 and 1.
        """
        proc_out = subprocess.run(
            ["geo2topo"],
            input=bytes(json.dumps(geojson), "utf-8"),
            stdout=subprocess.PIPE,
        )
        proc_out = subprocess.run(
            ["toposimplify", "-P", p],
            input=proc_out.stdout,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        topojson = json.loads(proc_out.stdout)
        # Standardize object name
        topojson["objects"]["divisions"] = topojson["objects"].pop("-")
        return topojson

    def get_district_shp(self):
        cmd_path = os.path.dirname(os.path.realpath(__file__))
        pa_path = os.path.join(cmd_path, "../../bin/pa")

        shape = shapefile.Reader(
            os.path.join(pa_path, "{}.shp".format(SHP_SLUG))
        )
        fields = shape.fields[1:]
        field_names = [f[0] for f in fields]
        district_records = [shp for shp in shape.shapeRecords()]
        features = []
        for shp in district_records:
            rec = dict(zip(field_names, shp.record))
            label = "{} congressional district".format(
                ordinal(int(rec["DISTRICT"]))
            )

            geometry = shp.shape.__geo_interface__
            geodata = {
                "type": "Feature",
                "geometry": geometry,
                "properties": {
                    "state": 42,
                    "district": rec["DISTRICT"],
                    "name": label,
                },
            }
            features.append(geodata)
        return self.toposimplify(geojson.FeatureCollection(features), "0.03")

    def handle(self, *args, **options):
        cmd_path = os.path.dirname(os.path.realpath(__file__))
        pa_path = os.path.join(cmd_path, "../../bin/pa")

        shape = shapefile.Reader(
            os.path.join(pa_path, "{}.shp".format(SHP_SLUG))
        )

        fields = shape.fields[1:]
        field_names = [f[0] for f in fields]
        districts = [shp for shp in shape.shapeRecords()]

        state = Division.objects.get(code=42, level__name=DivisionLevel.STATE)

        district_level = DivisionLevel.objects.get(name=DivisionLevel.DISTRICT)

        for shp in tqdm(districts):
            district = dict(zip(field_names, shp.record))
            label = "{} {} congressional district".format(
                state.label, ordinal(int(district["DISTRICT"]))
            )

            district_obj, created = Division.objects.update_or_create(
                code=district["DISTRICT"],
                level=district_level,
                parent=state,
                defaults={
                    "name": label,
                    "label": label,
                    "code_components": {
                        "fips": {"state": state.code},
                        "district": district["DISTRICT"],
                    },
                },
            )

            geodata = {
                "type": "Feature",
                "geometry": shp.shape.__geo_interface__,
                "properties": {
                    "state": 42,
                    "district": district["DISTRICT"],
                    "name": label,
                },
            }

            district_geometry = district_obj.geometries.get(
                division=district_obj,
                subdivision_level=district_level,
                series=2016,
            )

            district_geometry.topojson = self.toposimplify(geodata, "0.03")
            district_geometry.save()

        state_geometry = state.geometries.get(
            division=state, subdivision_level=district_level, series=2016
        )

        state_geometry.topojson = self.get_district_shp()
        state_geometry.save()



# Imports from geography.
from geography.models import DivisionLevel
from geography.serializers import DivisionLevelSerializer
from geography.viewsets.base import BaseViewSet


class DivisionLevelViewSet(BaseViewSet):
    queryset = DivisionLevel.objects.all()
    serializer_class = DivisionLevelSerializer



# Imports from python.
import os


# Imports from Django.
from django.core.management.base import BaseCommand


# Imports from other dependencies.
import geojson
import shapefile


# Imports from geography.
from geography.management.commands.bootstrap._arguments import ArgumentsMethods
from geography.management.commands.bootstrap._attributes import Attributes
from geography.management.commands.bootstrap._toposimplify import Toposimplify
from geography.models import Division
from geography.models import DivisionLevel
from geography.models import Geometry
from geography.utils.lookups import county_lookup


SHP_SLUG = "clipped_hi"
STATE_LEVEL = DivisionLevel.STATE


class Command(Toposimplify, ArgumentsMethods, Attributes, BaseCommand):
    def get_county_shp(self, fips):
        cmd_path = os.path.dirname(os.path.realpath(__file__))
        SHAPEFILE_PATH = os.path.join(cmd_path, "../../bin/hi")

        shape = shapefile.Reader(
            os.path.join(SHAPEFILE_PATH, "{}.shp".format(SHP_SLUG))
        )
        fields = shape.fields[1:]
        field_names = [f[0] for f in fields]

        county_records = [
            shp
            for shp in shape.shapeRecords()
            if dict(zip(field_names, shp.record))["STATEFP"] == fips
        ]
        features = []
        for shp in county_records:
            rec = dict(zip(field_names, shp.record))
            geometry = shp.shape.__geo_interface__
            geodata = {
                "type": "Feature",
                "geometry": geometry,
                "properties": {
                    "state": rec["STATEFP"],
                    "county": rec["COUNTYFP"],
                    "name": county_lookup[rec["STATEFP"]].get(
                        rec["COUNTYFP"], rec["NAME"]
                    ),
                },
            }
            features.append(geodata)
        threshold = (
            self.THRESHOLDS["nation"]
            if fips == "00"
            else self.THRESHOLDS["county"]
        )
        return self.toposimplify(
            geojson.FeatureCollection(features), threshold
        )

    def handle(self, *args, **options):
        self.set_attributes()
        self.parse_arguments(options)

        state = Division.objects.get(code=15, level__name=DivisionLevel.STATE)

        geojson, created = Geometry.objects.update_or_create(
            division=state,
            subdivision_level=self.COUNTY_LEVEL,
            simplification=0.075,
            source="https://www2.census.gov/geo/tiger/GENZ2016/shp/cb_2016_us_state_500k.zip",  # noqa
            series="2016",
            defaults={"topojson": self.get_county_shp("15")},
        )

        # TODO: District geometries

        self.stdout.write(self.style.SUCCESS("All done! 🏁"))



class ArgumentsMethods(object):
    def parse_arguments(self, options):
        self.YEAR = options["year"]
        self.STATES = options["states"]

    def add_arguments(self, parser):
        parser.add_argument(
            "states",
            nargs="+",
            help="States to export by FIPS code. \
            Use 00 to export all geographies.",
        )

        parser.add_argument(
            "--year",
            action="store",
            dest="year",
            default="2018",
            help="Specify year of shapefile series (default, 2018)",
        )



# Imports from other dependencies.
from rest_framework.pagination import PageNumberPagination


class ResultsPagination(PageNumberPagination):
    page_size = 10
    page_size_query_param = "page_size"
    max_page_size = 100



# Imports from python.
import os


# Imports from other dependencies.
import boto3


# Imports from geography.
from geography.conf import settings
from geography.models import DivisionLevel


def get_bucket():
    session = boto3.session.Session(
        region_name=settings.AWS_REGION,
        aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
        aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
    )
    s3 = session.resource("s3")
    return s3.Bucket(settings.AWS_S3_BUCKET)


class Attributes(object):
    def set_attributes(self):
        self.STATE_LEVEL = DivisionLevel.objects.get(name=DivisionLevel.STATE)
        self.DISTRICT_LEVEL = DivisionLevel.objects.get(
            name=DivisionLevel.DISTRICT
        )

        self.BUCKET = get_bucket()
        self.OUTPUT_PATH = os.path.join(
            settings.AWS_S3_UPLOAD_ROOT, "geography/us-census/cb/500k"
        )



# Generated by Django 2.0.1 on 2018-01-06 15:49

import django.contrib.postgres.fields.jsonb
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Division',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('uid', models.CharField(blank=True, editable=False, max_length=500)),
                ('slug', models.SlugField(blank=True, editable=False, max_length=255, unique=True)),
                ('name', models.CharField(max_length=255)),
                ('label', models.CharField(blank=True, max_length=255)),
                ('short_label', models.CharField(blank=True, max_length=50, null=True)),
                ('code', models.CharField(help_text='Code representing a geography: FIPS code for states and         counties, district number for districts, precinct number for         precincts, etc.', max_length=200)),
                ('code_components', django.contrib.postgres.fields.jsonb.JSONField(blank=True, help_text='Component parts of code', null=True)),
                ('effective', models.BooleanField(default=True)),
                ('effective_start', models.DateTimeField(blank=True, null=True)),
                ('effective_end', models.DateTimeField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='DivisionLevel',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('uid', models.CharField(blank=True, editable=False, max_length=500)),
                ('slug', models.SlugField(blank=True, editable=False, max_length=255, unique=True)),
                ('name', models.CharField(choices=[('country', 'Country'), ('state', 'State'), ('district', 'District'), ('county', 'County'), ('township', 'Township'), ('precinct', 'Precinct')], max_length=255, unique=True)),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='children', to='geography.DivisionLevel')),
            ],
        ),
        migrations.CreateModel(
            name='Geometry',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('simplification', models.FloatField(help_text='Minimum quantile of planar         triangle areas for simplfying topojson.', validators=[django.core.validators.MinValueValidator(0.0), django.core.validators.MaxValueValidator(1.0)])),
                ('topojson', django.contrib.postgres.fields.jsonb.JSONField()),
                ('source', models.URLField(blank=True, help_text='Link to the source of this geography data.', null=True)),
                ('series', models.CharField(blank=True, help_text='Year of boundary series, e.g., 2016 TIGER/Line files.', max_length=4, null=True)),
                ('effective', models.BooleanField(default=True)),
                ('effective_start', models.DateField(blank=True, null=True)),
                ('effective_end', models.DateField(blank=True, null=True)),
                ('division', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='geometries', to='geography.Division')),
                ('subdivision_level', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='+', to='geography.DivisionLevel')),
            ],
            options={
                'verbose_name_plural': 'Geometries',
            },
        ),
        migrations.CreateModel(
            name='IntersectRelationship',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('intersection', models.DecimalField(blank=True, decimal_places=6, help_text='The portion of the to_division that intersects this division.', max_digits=7, null=True)),
                ('from_division', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='geography.Division')),
                ('to_division', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='geography.Division')),
            ],
        ),
        migrations.AddField(
            model_name='division',
            name='intersecting',
            field=models.ManyToManyField(help_text='Intersecting divisions intersect this one geographically but do not necessarily have a parent/child relationship. The relationship between a congressional district and a precinct is an example of an intersecting relationship.', related_name='_division_intersecting_+', through='geography.IntersectRelationship', to='geography.Division'),
        ),
        migrations.AddField(
            model_name='division',
            name='level',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='divisions', to='geography.DivisionLevel'),
        ),
        migrations.AddField(
            model_name='division',
            name='parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='children', to='geography.Division'),
        ),
        migrations.AlterUniqueTogether(
            name='intersectrelationship',
            unique_together={('from_division', 'to_division')},
        ),
    ]



# Imports from geography.
from geography.serializers.division import DivisionSerializer
from geography.serializers.division_level import DivisionLevelSerializer
from geography.serializers.geometry import GeometrySerializer
from geography.serializers.geometry_list import GeometryListSerializer
from geography.serializers.point_label_offset import PointLabelOffsetSerializer
from geography.serializers.point import PointSerializer



# Imports from Django.
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import ugettext_lazy as _


# Imports from other dependencies.
from civic_utils.models import CivicBaseModel


class IntersectRelationship(CivicBaseModel):
    """Part of a paired relationship between intersecting divisions.

    Each IntersectRelationship instance represents one side of such a
    paired relationship.

    The intersection field represents the decimal proportion of the
    to_division that intersects with the from_division. It's useful for
    apportioning counts between the areas, for example, population
    statistics from census data.
    """

    natural_key_fields = ["from_division", "to_division"]

    from_division = models.ForeignKey(
        "Division", on_delete=models.CASCADE, related_name="+"
    )
    to_division = models.ForeignKey(
        "Division", on_delete=models.CASCADE, related_name="+"
    )
    intersection = models.DecimalField(
        max_digits=7,
        decimal_places=6,
        null=True,
        blank=True,
        help_text="The portion of the to_division that intersects this "
        "division.",
    )

    class Meta:
        # Don't allow duplicate relationships between divisions
        unique_together = ("from_division", "to_division")

    def clean(self):
        if self.intersection < 0.0 or self.intersection > 1.0:
            raise ValidationError(
                _("Intersection should be a decimal between 0 and 1.")
            )



# Generated by Django 2.2.4 on 2019-08-21 21:22

import datetime
import django.core.validators
from django.db import migrations, models
from django.utils import timezone


class Migration(migrations.Migration):

    dependencies = [("geography", "0002_point_pointlabeloffset")]

    operations = [
        migrations.AddField(
            model_name="division",
            name="created",
            field=models.DateTimeField(
                default=timezone.utc.localize(
                    datetime.datetime(2019, 1, 1, 0, 0)
                ),
                editable=False,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="division",
            name="updated",
            field=models.DateTimeField(
                default=timezone.utc.localize(
                    datetime.datetime(2019, 1, 1, 0, 0)
                ),
                editable=False,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="divisionlevel",
            name="created",
            field=models.DateTimeField(
                default=timezone.utc.localize(
                    datetime.datetime(2019, 1, 1, 0, 0)
                ),
                editable=False,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="divisionlevel",
            name="updated",
            field=models.DateTimeField(
                default=timezone.utc.localize(
                    datetime.datetime(2019, 1, 1, 0, 0)
                ),
                editable=False,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="geometry",
            name="created",
            field=models.DateTimeField(
                default=timezone.utc.localize(
                    datetime.datetime(2019, 1, 1, 0, 0)
                ),
                editable=False,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="geometry",
            name="data_summary",
            field=models.SlugField(
                blank=True,
                editable=False,
                help_text="A brief description (written as a slug) of what makes this data unique from other Geometry instances with the same 'division' and 'subdivision_level' values.",
                max_length=255,
                unique=True,
            ),
        ),
        migrations.AddField(
            model_name="geometry",
            name="uid",
            field=models.CharField(blank=True, editable=False, max_length=500),
        ),
        migrations.AddField(
            model_name="geometry",
            name="updated",
            field=models.DateTimeField(
                default=timezone.utc.localize(
                    datetime.datetime(2019, 1, 1, 0, 0)
                ),
                editable=False,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="intersectrelationship",
            name="created",
            field=models.DateTimeField(
                default=timezone.utc.localize(
                    datetime.datetime(2019, 1, 1, 0, 0)
                ),
                editable=False,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="intersectrelationship",
            name="updated",
            field=models.DateTimeField(
                default=timezone.utc.localize(
                    datetime.datetime(2019, 1, 1, 0, 0)
                ),
                editable=False,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="point",
            name="created",
            field=models.DateTimeField(
                default=timezone.utc.localize(
                    datetime.datetime(2019, 1, 1, 0, 0)
                ),
                editable=False,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="point",
            name="uid",
            field=models.CharField(blank=True, editable=False, max_length=500),
        ),
        migrations.AddField(
            model_name="point",
            name="updated",
            field=models.DateTimeField(
                default=timezone.utc.localize(
                    datetime.datetime(2019, 1, 1, 0, 0)
                ),
                editable=False,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="pointlabeloffset",
            name="created",
            field=models.DateTimeField(
                default=timezone.utc.localize(
                    datetime.datetime(2019, 1, 1, 0, 0)
                ),
                editable=False,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="pointlabeloffset",
            name="uid",
            field=models.CharField(blank=True, editable=False, max_length=500),
        ),
        migrations.AddField(
            model_name="pointlabeloffset",
            name="updated",
            field=models.DateTimeField(
                default=timezone.utc.localize(
                    datetime.datetime(2019, 1, 1, 0, 0)
                ),
                editable=False,
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="geometry",
            name="simplification",
            field=models.FloatField(
                help_text="Minimum quantile of planar triangle areas for simplfying topojson.",
                validators=[
                    django.core.validators.MinValueValidator(0.0),
                    django.core.validators.MaxValueValidator(1.0),
                ],
            ),
        ),
        migrations.AlterUniqueTogether(
            name="geometry",
            unique_together={
                ("division", "subdivision_level", "data_summary")
            },
        ),
        migrations.AlterUniqueTogether(
            name="point", unique_together={("geometry", "label", "lat", "lon")}
        ),
    ]



# Imports from python.
import json
import os


# Imports from other dependencies.
import shapefile
from tqdm import tqdm
import us
from uuslug import slugify


# Imports from geography.
from geography.models import Division
from geography.models import Geometry
from geography.utils.lookups import township_states


class StateFixtures(object):
    def create_state_fixtures(self):
        SHP_SLUG = "cb_{}_us_state_500k".format(self.YEAR)
        DOWNLOAD_PATH = os.path.join(self.DOWNLOAD_DIRECTORY, SHP_SLUG)

        shape = shapefile.Reader(
            os.path.join(DOWNLOAD_PATH, "{}.shp".format(SHP_SLUG))
        )
        fields = shape.fields[1:]
        field_names = [f[0] for f in fields]

        nation_obj = Division.objects.get(code="00", level=self.NATIONAL_LEVEL)

        for shp in tqdm(shape.shapeRecords(), desc="States"):
            state = dict(zip(field_names, shp.record))
            postal = us.states.lookup(state["STATEFP"]).abbr

            state_obj, created = Division.objects.update_or_create(
                code=state["STATEFP"],
                level=self.STATE_LEVEL,
                parent=nation_obj,
                defaults={
                    "name": state["NAME"],
                    "label": state["NAME"],
                    "code_components": {
                        "fips": {"state": state["STATEFP"]},
                        "postal": postal,
                    },
                },
            )
            geodata = {
                "type": "Feature",
                "geometry": shp.shape.__geo_interface__,
                "properties": {
                    "state": state["STATEFP"],
                    "name": state["NAME"],
                },
            }
            geojson, created = Geometry.objects.update_or_create(
                division=state_obj,
                subdivision_level=self.STATE_LEVEL,
                simplification=self.THRESHOLDS["state"],
                data_summary=slugify(
                    "{}--{}--{}".format(
                        state_obj.slug,
                        self.STATE_LEVEL.slug,
                        self.THRESHOLDS["state"],
                    )
                ),
                source=os.path.join(
                    self.SHP_SOURCE_BASE.format(self.YEAR), SHP_SLUG
                )
                + ".zip",
                series=self.YEAR,
                defaults={
                    "topojson": self.toposimplify(
                        geodata, self.THRESHOLDS["state"]
                    )
                },
            )
            geojson, created = Geometry.objects.update_or_create(
                division=state_obj,
                subdivision_level=self.COUNTY_LEVEL,
                simplification=self.THRESHOLDS["county"],
                data_summary=slugify(
                    "{}--{}--{}".format(
                        state_obj.slug,
                        self.COUNTY_LEVEL.slug,
                        self.THRESHOLDS["county"],
                    )
                ),
                source=os.path.join(
                    self.SHP_SOURCE_BASE.format(self.YEAR), SHP_SLUG
                )
                + ".zip",
                series=self.YEAR,
                defaults={
                    "topojson": self.get_state_county_shps(state["STATEFP"])
                },
            )
            geojson, created = Geometry.objects.update_or_create(
                division=state_obj,
                subdivision_level=self.DISTRICT_LEVEL,
                simplification=self.THRESHOLDS["district"],
                data_summary=slugify(
                    "{}--{}--{}".format(
                        state_obj.slug,
                        self.DISTRICT_LEVEL.slug,
                        self.THRESHOLDS["district"],
                    )
                ),
                source=os.path.join(
                    self.SHP_SOURCE_BASE.format(self.YEAR),
                    "cb_{}_us_cd{}_500k".format(self.YEAR, self.CONGRESS),
                )
                + ".zip",
                series=self.YEAR,
                defaults={
                    "topojson": self.get_state_district_shps(state["STATEFP"])
                },
            )

            if postal in township_states:
                geojson, created = Geometry.objects.update_or_create(
                    division=state_obj,
                    subdivision_level=self.TOWNSHIP_LEVEL,
                    simplification=self.THRESHOLDS["county"],
                    data_summary=slugify(
                        "{}--{}--{}".format(
                            state_obj.slug,
                            self.TOWNSHIP_LEVEL.slug,
                            self.THRESHOLDS["county"],
                        )
                    ),
                    source=os.path.join(
                        self.SHP_SOURCE_BASE.format(self.YEAR),
                        "cb_{}_{}_cousub_500k".format(
                            self.YEAR, state["STATEFP"]
                        ),
                    )
                    + ".zip",
                    series=self.YEAR,
                    defaults={
                        "topojson": self.get_state_township_shps(
                            state["STATEFP"]
                        )
                    },
                )

            tqdm.write(
                self.TQDM_PREFIX
                + ">  FIPS {}  @ ~{}kb     ".format(
                    state["STATEFP"],
                    round(len(json.dumps(geojson.topojson)) / 1000),
                )
            )
        tqdm.write(self.style.SUCCESS("Done.\n"))



# flake8: noqa
from .division import DivisionViewSet
from .geometry import GeometryViewSet
from .division_level import DivisionLevelViewSet
from .point_label_offset import PointLabelOffsetViewSet
from .point import PointViewSet



# Imports from Django.
from django.conf import settings as project_settings


class Settings:
    pass


Settings.CENSUS_API_KEY = getattr(project_settings, "CENSUS_API_KEY", None)

Settings.AWS_ACCESS_KEY_ID = getattr(
    project_settings, "GEOGRAPHY_AWS_ACCESS_KEY_ID", None
)

Settings.AWS_SECRET_ACCESS_KEY = getattr(
    project_settings, "GEOGRAPHY_AWS_SECRET_ACCESS_KEY", None
)

Settings.AWS_REGION = getattr(
    project_settings, "GEOGRAPHY_AWS_REGION", "us-east-1"
)

Settings.AWS_S3_BUCKET = getattr(
    project_settings, "GEOGRAPHY_AWS_S3_BUCKET", None
)

Settings.AWS_S3_UPLOAD_ROOT = getattr(
    project_settings, "GEOGRAPHY_AWS_S3_UPLOAD_ROOT", "elections"
)

Settings.AWS_ACL = getattr(
    project_settings, "GEOGRAPHY_AWS_ACL", "public-read"
)

Settings.AWS_CACHE_HEADER = getattr(
    project_settings, "GEOGRAPHY_AWS_CACHE_HEADER", "max-age=3600"
)

Settings.API_AUTHENTICATION_CLASS = getattr(
    project_settings,
    "GEOGRAPHY_API_AUTHENTICATION_CLASS",
    "rest_framework.authentication.BasicAuthentication",
)

Settings.API_PERMISSION_CLASS = getattr(
    project_settings,
    "GEOGRAPHY_API_PERMISSION_CLASS",
    "rest_framework.permissions.IsAdminUser",
)

Settings.API_PAGINATION_CLASS = getattr(
    project_settings,
    "GEOGRAPHY_API_PAGINATION_CLASS",
    "geography.pagination.ResultsPagination",
)

settings = Settings



# Imports from other dependencies.
from civic_utils.serializers import CommandLineListSerializer
from civic_utils.serializers import NaturalKeySerializerMixin


# Imports from geography.
from geography.models import PointLabelOffset


class PointLabelOffsetSerializer(
    NaturalKeySerializerMixin, CommandLineListSerializer
):
    class Meta(CommandLineListSerializer.Meta):
        model = PointLabelOffset
        fields = "__all__"



# Imports from geography.
from geography.management.commands.bake.geometries._state import (
    StateGeometries,
)
from geography.management.commands.bake.geometries._district import (
    DistrictGeometries,
)



# Generated by Django 2.2.4 on 2020-03-09 23:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geography', '0003_auto_20190821_2122'),
    ]

    operations = [
        migrations.AlterField(
            model_name='divisionlevel',
            name='name',
            field=models.CharField(choices=[('country', 'Country'), ('state', 'State'), ('district', 'District'), ('county', 'County'), ('township', 'Township'), ('precinct', 'Precinct'), ('abroad', 'Voters abroad')], max_length=255, unique=True),
        ),
    ]



# Imports from Django.
from django.contrib.postgres.fields import JSONField
from django.db import models


# Imports from other dependencies.
from civic_utils.models import CivicBaseModel
from civic_utils.models import UniqueIdentifierMixin


class Point(UniqueIdentifierMixin, CivicBaseModel):
    """A point is a city."""

    natural_key_fields = ["geometry", "label", "lat", "lon"]
    uid_prefix = "point"

    geometry = models.ForeignKey(
        "Geometry", on_delete=models.CASCADE, related_name="points"
    )

    lat = models.FloatField(
        help_text="Latitude coordinate in decimal degrees."
    )
    lon = models.FloatField(
        help_text="Longitude coordinate in decimal degrees."
    )

    attributes = JSONField(
        blank=True,
        null=True,
        help_text="Miscellaneous attributes on the point.",
    )

    threshold = models.PositiveSmallIntegerField(
        default=0,
        help_text="A threshold in pixels above which to display this point.",
    )

    label = models.CharField(max_length=250)

    class Meta:
        unique_together = ("geometry", "label", "lat", "lon")

    def __str__(self):
        return "{}, {}".format(self.label, self.geometry.division.name)

    def get_uid_suffix(self):
        return (
            f"geometry={self.geometry.uid}"
            f"&label={self.label}"
            f"&coords={self.lat},{self.lon}"
        )

    def to_topojson(self):
        offsets = (
            [offset.to_object() for offset in self.offsets.all()]
            if len(self.offsets.all()) > 0
            else None
        )
        return {
            "type": "Point",
            "coordinates": [self.lon, self.lat],
            "properties": {
                "label": self.label,
                "threshold": self.threshold,
                "attributes": self.attributes,
                "offsets": offsets,
            },
        }


class PointLabelOffset(UniqueIdentifierMixin, CivicBaseModel):
    """Offsets used to display a Point's label."""

    natural_key_fields = ["point", "threshold"]
    uid_prefix = "pointlabel"

    point = models.ForeignKey(
        Point, on_delete=models.CASCADE, related_name="offsets"
    )
    x = models.SmallIntegerField(
        default=0, help_text="Lateral offset in pixels."
    )
    y = models.SmallIntegerField(
        default=0, help_text="Vertical offset in pixels."
    )
    threshold = models.PositiveSmallIntegerField(
        default=0,
        help_text="A threshold in pixels above which to apply this offset.",
    )

    def __str__(self):
        return "{} @ {}".format(self.threshold, self.point)

    def get_uid_suffix(self):
        return f"point={self.point.uid}&threshold={self.threshold}"

    def to_object(self):
        return {"threshold": self.threshold, "x": self.x, "y": self.y}



# Imports from python.
import os


# Imports from other dependencies.
import geojson
import shapefile


# Imports from geography.
from geography.utils.lookups import county_lookup


class StateCountyShapes(object):
    def get_state_county_shps(self, fips):
        SHP_SLUG = "cb_{}_us_county_500k".format(self.YEAR)
        DOWNLOAD_PATH = os.path.join(self.DOWNLOAD_DIRECTORY, SHP_SLUG)
        shape = shapefile.Reader(
            os.path.join(DOWNLOAD_PATH, "{}.shp".format(SHP_SLUG))
        )
        fields = shape.fields[1:]
        field_names = [f[0] for f in fields]
        county_records = [
            shp
            for shp in shape.shapeRecords()
            if dict(zip(field_names, shp.record))["STATEFP"] == fips
            # National county map
            or (
                fips == "00"
                and int(dict(zip(field_names, shp.record))["STATEFP"]) <= 56
            )
        ]
        features = []
        for shp in county_records:
            rec = dict(zip(field_names, shp.record))

            # Skip creating counties for territories and DC.
            if int(rec["STATEFP"]) > 56 or int(rec["STATEFP"]) == 11:
                continue

            geometry = shp.shape.__geo_interface__
            geodata = {
                "type": "Feature",
                "geometry": geometry,
                "properties": {
                    "state": rec["STATEFP"],
                    "county": rec["COUNTYFP"],
                    "name": county_lookup[rec["STATEFP"]].get(
                        rec["COUNTYFP"], rec["NAME"]
                    ),
                },
            }
            features.append(geodata)
        threshold = (
            self.THRESHOLDS["nation"]
            if fips == "00"
            else self.THRESHOLDS["county"]
        )
        return self.toposimplify(
            geojson.FeatureCollection(features), threshold
        )



# Imports from python.
import os
from pathlib import Path
import shutil
import urllib.request as request
import zipfile


class DownloadState(object):
    def download_state_data(self):
        SHP_SLUG = "cb_{}_us_state_500k".format(self.YEAR)
        DOWNLOAD_PATH = os.path.join(self.DOWNLOAD_DIRECTORY, SHP_SLUG)
        ZIPFILE = "{}{}.zip".format(DOWNLOAD_PATH, SHP_SLUG)
        SHP_PATH = os.path.join(
            self.SHP_SOURCE_BASE.format(self.YEAR), SHP_SLUG
        )

        if not os.path.exists(DOWNLOAD_PATH):
            os.makedirs(DOWNLOAD_PATH)

        if not Path(ZIPFILE).is_file():
            with request.urlopen("{}.zip".format(SHP_PATH)) as response, open(
                ZIPFILE, "wb"
            ) as out_file:
                shutil.copyfileobj(response, out_file)

        if not Path("{}{}.shp".format(DOWNLOAD_PATH, SHP_SLUG)).is_file():
            with zipfile.ZipFile(ZIPFILE, "r") as file:
                file.extractall(DOWNLOAD_PATH)



# Imports from python.
import os


# Imports from other dependencies.
import geojson
import shapefile
from shapely.geometry import mapping
from shapely.geometry import shape


# Imports from geography.
from geography.utils.lookups import county_lookup


class DistrictCountyShapes(object):
    def get_district_county_shps(self, fips, district_code):
        DISTRICT_SHP_SLUG = "cb_{}_us_cd{}_500k".format(
            self.YEAR, self.CONGRESS
        )
        DISTRICT_SHP_PATH = os.path.join(
            self.DOWNLOAD_DIRECTORY, DISTRICT_SHP_SLUG
        )
        district_shape = shapefile.Reader(
            os.path.join(DISTRICT_SHP_PATH, "{}.shp".format(DISTRICT_SHP_SLUG))
        )
        district_fields = district_shape.fields[1:]
        district_field_names = [f[0] for f in district_fields]
        district_code_key = "CD{}FP".format(self.CONGRESS)
        district_record = [
            shp
            for shp in district_shape.shapeRecords()
            if dict(zip(district_field_names, shp.record))["STATEFP"] == fips
            and dict(zip(district_field_names, shp.record))[district_code_key]
            == district_code
        ][0]
        district_shapely = shape(district_record.shape.__geo_interface__)

        COUNTY_SHP_SLUG = "cb_{}_us_county_500k".format(self.YEAR)
        COUNTY_SHP_PATH = os.path.join(
            self.DOWNLOAD_DIRECTORY, COUNTY_SHP_SLUG
        )

        county_shape = shapefile.Reader(
            os.path.join(COUNTY_SHP_PATH, "{}.shp".format(COUNTY_SHP_SLUG))
        )
        county_fields = county_shape.fields[1:]
        county_field_names = [f[0] for f in county_fields]
        county_records = [
            shp
            for shp in county_shape.shapeRecords()
            if dict(zip(county_field_names, shp.record))["STATEFP"] == fips
            and shape(shp.shape.__geo_interface__).intersects(district_shapely)
        ]
        features = []
        for shp in county_records:
            intersection = shape(shp.shape.__geo_interface__).intersection(
                district_shapely
            )
            rec = dict(zip(county_field_names, shp.record))
            geodata = {
                "type": "Feature",
                "geometry": mapping(intersection),
                "properties": {
                    "state": rec["STATEFP"],
                    "district": district_code,
                    "county": rec["COUNTYFP"],
                    "name": county_lookup[rec["STATEFP"]].get(
                        rec["COUNTYFP"], rec["NAME"]
                    ),
                },
            }
            features.append(geodata)
        return self.toposimplify(
            geojson.FeatureCollection(features), self.THRESHOLDS["county"]
        )



# Imports from Django.
from django.contrib import admin


# Imports from geography.
from geography.models import Division
from geography.models import DivisionLevel
from geography.models import Geometry
from geography.models import IntersectRelationship
from geography.models import Point
from geography.models import PointLabelOffset


def custom_titled_filter(title):
    class Wrapper(admin.FieldListFilter):
        def __new__(cls, *args, **kwargs):
            instance = admin.FieldListFilter.create(*args, **kwargs)
            instance.title = title
            return instance

    return Wrapper


class IntersectRelationshipInline(admin.StackedInline):
    model = IntersectRelationship
    fk_name = "from_division"
    extra = 0


class DivisionAdmin(admin.ModelAdmin):
    inlines = (IntersectRelationshipInline,)
    list_display = ("label", "level", "code")
    list_filter = ("level",)
    search_fields = ("code", "label")
    readonly_fields = ("parent", "uid")

    fieldsets = (
        ("Names", {"fields": ("name", "label", "short_label")}),
        ("Reference codes", {"fields": ("code", "code_components")}),
        (
            "In effect",
            {"fields": ("effective", "effective_start", "effective_end")},
        ),
        ("Reference fields", {"fields": ("parent", "uid")}),
    )


class PointInline(admin.StackedInline):
    model = Point
    fk_name = "geometry"
    extra = 0


class GeometryAdmin(admin.ModelAdmin):
    inlines = (PointInline,)
    list_display = ("division", "small_preview", "map_level", "series")
    list_filter = (
        "series",
        ("division__level", custom_titled_filter("division level")),
        ("subdivision_level", custom_titled_filter("map level")),
    )
    search_fields = ("division__name",)
    readonly_fields = ("file_size", "large_preview", "source", "series")

    fieldsets = (
        (
            None,
            {"fields": ("division", "subdivision_level", "source", "series")},
        ),
        (
            "Geo data",
            {
                "fields": (
                    "topojson",
                    "simplification",
                    "file_size",
                    "large_preview",
                )
            },
        ),
        (
            "In effect",
            {"fields": ("effective", "effective_start", "effective_end")},
        ),
    )

    def map_level(self, obj):
        return obj.subdivision_level.name


class PointLabelOffsetInline(admin.StackedInline):
    model = PointLabelOffset
    fk_name = "point"
    extra = 0


class PointAdmin(admin.ModelAdmin):
    inlines = (PointLabelOffsetInline,)


admin.site.register(DivisionLevel)
admin.site.register(Division, DivisionAdmin)
admin.site.register(Geometry, GeometryAdmin)
admin.site.register(Point, PointAdmin)



# Imports from geography.
from geography.models import Division
from geography.serializers import DivisionSerializer
from geography.viewsets.base import BaseViewSet


class DivisionViewSet(BaseViewSet):
    queryset = Division.objects.all()
    serializer_class = DivisionSerializer



# Imports from Django.
from django.core.management.base import CommandError


class ArgumentsMethods(object):
    def parse_arguments(self, options):
        self.YEAR = options["year"]
        self.CONGRESS = options["congress"]
        self.THRESHOLDS = {
            "nation": str(options["nationThreshold"]),
            "state": str(options["stateThreshold"]),
            "district": str(options["districtThreshold"]),
            "county": str(options["countyThreshold"]),
        }

    def add_arguments(self, parser):
        def check_threshold(arg):
            value = float(arg)
            if value < 0 or value > 1:
                raise CommandError(
                    "Threshold must be a decimal between 0 and 1."
                )
            return value

        parser.add_argument(
            "--year",
            action="store",
            dest="year",
            default="2018",
            help="Specify year of shapefile series (default, 2018)",
        )
        parser.add_argument(
            "--congress",
            action="store",
            dest="congress",
            default="116",
            help="Specify congress of district shapefile series \
             (default, 116)",
        )
        parser.add_argument(
            "--nationThreshold",
            type=check_threshold,
            default=0.005,
            dest="nationThreshold",
            help="Simplification threshold value for nation topojson \
                (default, 0.005)",
        )
        parser.add_argument(
            "--stateThreshold",
            type=check_threshold,
            default=0.05,
            dest="stateThreshold",
            help="Simplification threshold value for state topojson \
                (default, 0.05)",
        )
        parser.add_argument(
            "--districtThreshold",
            type=check_threshold,
            default=0.08,
            dest="districtThreshold",
            help="Simplification threshold value for district topojson \
                (default, 0.08)",
        )
        parser.add_argument(
            "--countyThreshold",
            type=check_threshold,
            default=0.075,
            dest="countyThreshold",
            help="Simplification threshold value for county topojson \
                (default, 0.075)",
        )



# Imports from geography.
from geography.management.commands.bootstrap.aggregate_shapes.state._county import (
    StateCountyShapes,
)
from geography.management.commands.bootstrap.aggregate_shapes.state._district import (
    StateDistrictShapes,
)
from geography.management.commands.bootstrap.aggregate_shapes.state._township import (
    StateTownshipShapes,
)



# Imports from python.
import json
import subprocess


class Toposimplify:
    @staticmethod
    def toposimplify(geojson, p):
        """
        Convert geojson and simplify topology.

        geojson is a dict representing geojson.
        p is a simplification threshold value between 0 and 1.
        """
        proc_out = subprocess.run(
            ["geo2topo"],
            input=bytes(json.dumps(geojson), "utf-8"),
            stdout=subprocess.PIPE,
        )
        proc_out = subprocess.run(
            ["toposimplify", "-P", p],
            input=proc_out.stdout,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        topojson = json.loads(proc_out.stdout)
        # Standardize object name
        topojson["objects"]["divisions"] = topojson["objects"].pop("-")
        return topojson



# Imports from geography.
from geography.management.commands.bootstrap.download._county import (
    DownloadCounty,
)
from geography.management.commands.bootstrap.download._district import (
    DownloadDistrict,
)
from geography.management.commands.bootstrap.download._township import (
    DownloadTownship,
)
from geography.management.commands.bootstrap.download._state import (
    DownloadState,
)



# Imports from other dependencies.
from census import Census
from DictObject import DictObject
import us
from tqdm import tqdm

# Imports from geography.
from geography.conf import settings


tqdm.write("📌  Loading U.S. Census geo lookups...")

census = Census(settings.CENSUS_API_KEY)

############
# COUNTIES
tqdm.write("   County lookups")
counties = census.sf1.get("NAME", geo={"for": "county:*"})
county_lookup = DictObject({})


for county in tqdm(counties):
    c = DictObject(county)
    if c.state not in county_lookup:
        county_lookup[c.state] = {}
    county_lookup[c.state][c.county] = c.NAME

#############
# TOWNSHIPS
tqdm.write("   State township lookups")
townships = []
township_states = ["CT", "MA", "ME", "NH", "RI", "VT"]
township_lookup = DictObject({})
for state in tqdm(township_states):
    state_codes = us.states.lookup(state)

    state_townships = census.sf1.get(
        "NAME",
        geo={
            "for": "county subdivision:*",
            "in": "state:{} county:*".format(state_codes.fips),
        },
    )
    townships.extend(state_townships)
for township in townships:
    t = DictObject(township)
    if t.state not in township_lookup:
        township_lookup[t.state] = {}

    if t.county not in township_lookup[t.state]:
        township_lookup[t.state][t.county] = {}

    township_lookup[t.state][t.county][t.county_subdivision] = t.NAME



# Imports from python.
import os
from setuptools import find_packages
from setuptools import setup


# Imports from geography.
from geography import __version__


REPO_URL = "https://github.com/The-Politico/politico-civic-geography/"

PYPI_VERSION = ".".join(str(v) for v in __version__)

with open(os.path.join(os.path.dirname(__file__), "README.md")) as readme:
    README = readme.read()


setup(
    name="politico-civic-geography",
    version=PYPI_VERSION,
    packages=find_packages(exclude=["docs", "tests", "example"]),
    include_package_data=True,
    license="MIT",
    description=(
        "Manage political geographic and spatial data, the POLITICO way."
    ),
    long_description=README,
    long_description_content_type="text/markdown",
    url=REPO_URL,
    author="POLITICO interactive news",
    author_email="interactives@politico.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP",
    ],
    keywords="",
    install_requires=[
        "boto3",
        "census",
        "DictObject",
        "django",
        "djangorestframework",
        "dj-database-url",
        "geojson",
        "psycopg2-binary",
        "politico-civic-utils",
        "pyshp",
        "shapely",
        "stringcase",
        "tqdm",
        "us",
    ],
    extras_require={
        "dev": ["sphinx", "sphinxcontrib-django", "sphinx-rtd-theme"],
        "test": ["pytest"],
    },
)



# Imports from python.


# Imports from Django.
from django.urls import include, path


# Imports from other dependencies.
from rest_framework import routers


# Imports from geography.
from geography.viewsets import DivisionLevelViewSet
from geography.viewsets import DivisionViewSet
from geography.viewsets import GeometryViewSet
from geography.viewsets import PointLabelOffsetViewSet
from geography.viewsets import PointViewSet


router = routers.DefaultRouter()


router.register(r"divisions", DivisionViewSet)
router.register(r"geometries", GeometryViewSet)
router.register(r"division-levels", DivisionLevelViewSet)
router.register(r"point-label-offsets", PointLabelOffsetViewSet)
router.register(r"points", PointViewSet)


urlpatterns = [path("api/", include(router.urls))]



# Imports from python.
import json
import os


# Imports from other dependencies.
import geojson
import shapefile
from tqdm import tqdm
from uuslug import slugify


# Imports from geography.
from geography.models import Geometry


class NationFixtures(object):
    def create_nation_fixtures(self):
        """
        Create national US and State Map
        """
        SHP_SLUG = "cb_{}_us_state_500k".format(self.YEAR)
        DOWNLOAD_PATH = os.path.join(self.DOWNLOAD_DIRECTORY, SHP_SLUG)

        shape = shapefile.Reader(
            os.path.join(DOWNLOAD_PATH, "{}.shp".format(SHP_SLUG))
        )
        fields = shape.fields[1:]
        field_names = [f[0] for f in fields]
        features = []
        for shp in shape.shapeRecords():
            state = dict(zip(field_names, shp.record))
            geodata = {
                "type": "Feature",
                "geometry": shp.shape.__geo_interface__,
                "properties": {
                    "state": state["STATEFP"],
                    "name": state["NAME"],
                },
            }
            features.append(geodata)
        Geometry.objects.update_or_create(
            division=self.NATION,
            subdivision_level=self.STATE_LEVEL,
            simplification=self.THRESHOLDS["nation"],
            data_summary=slugify(
                "{}--{}--{}".format(
                    self.NATION.slug,
                    self.STATE_LEVEL.slug,
                    self.THRESHOLDS["nation"],
                )
            ),
            source=os.path.join(
                self.SHP_SOURCE_BASE.format(self.YEAR), SHP_SLUG
            )
            + ".zip",
            series=self.YEAR,
            defaults={
                "topojson": self.toposimplify(
                    geojson.FeatureCollection(features),
                    self.THRESHOLDS["nation"],
                )
            },
        )

        geo, created = Geometry.objects.update_or_create(
            division=self.NATION,
            subdivision_level=self.COUNTY_LEVEL,
            simplification=self.THRESHOLDS["nation"],
            data_summary=slugify(
                "{}--{}--{}".format(
                    self.NATION.slug,
                    self.COUNTY_LEVEL.slug,
                    self.THRESHOLDS["nation"],
                )
            ),
            source=os.path.join(
                self.SHP_SOURCE_BASE.format(self.YEAR), SHP_SLUG
            )
            + ".zip",
            series=self.YEAR,
            defaults={"topojson": self.get_state_county_shps("00")},
        )
        tqdm.write("Nation\n")
        tqdm.write(
            self.TQDM_PREFIX
            + ">  FIPS {}  @ ~{}kb     ".format(
                "00", round(len(json.dumps(geo.topojson)) / 1000)
            )
        )
        tqdm.write(self.style.SUCCESS("Done.\n"))



# Imports from python.
import os


# Imports from other dependencies.
from tqdm import tqdm


# Imports from geography.
from geography.conf import settings
from geography.models import Division
from geography.models import Geometry


class StateGeometries(object):
    def bake_state_geometries(self):
        state_geometries = None
        if self.STATES[0] == "00":
            state_geometries = Geometry.objects.filter(
                series=self.YEAR, division__level=self.STATE_LEVEL
            )
        else:
            for state in self.STATES:
                division = Division.objects.get(
                    level=self.STATE_LEVEL, code=state
                )
                if not state_geometries:
                    state_geometries = division.geometries.filter(
                        series=self.YEAR
                    )
                else:
                    state_geometries = (
                        state_geometries
                        | division.geometries.filter(series=self.YEAR)
                    )

        for geometry in tqdm(state_geometries, desc="State Geometries"):
            key = os.path.join(
                self.OUTPUT_PATH,
                self.YEAR,
                "states",
                geometry.division.code,
                "{0}-{1}.json".format(
                    geometry.division.code, geometry.subdivision_level.slug
                ),
            )
            self.BUCKET.put_object(
                Key=key,
                ACL=settings.AWS_ACL,
                Body=geometry.to_topojson(),
                CacheControl=settings.AWS_CACHE_HEADER,
                ContentType="application/json",
            )
        tqdm.write(self.style.SUCCESS("Done.\n"))



# Imports from python.
import os
from pathlib import Path
import shutil
import urllib.request as request
import zipfile


# Imports from other dependencies.
import us


# Imports from geography.
from geography.utils.lookups import township_states


class DownloadTownship(object):
    def download_township_data(self):
        for state in township_states:
            state_fips = us.states.lookup(state).fips
            SHP_SLUG = "cb_{}_{}_cousub_500k".format(self.YEAR, state_fips)
            DOWNLOAD_PATH = os.path.join(self.DOWNLOAD_DIRECTORY, SHP_SLUG)
            ZIPFILE = "{}{}.zip".format(DOWNLOAD_PATH, SHP_SLUG)
            SHP_PATH = os.path.join(
                self.SHP_SOURCE_BASE.format(self.YEAR), SHP_SLUG
            )
            if not Path(ZIPFILE).is_file():
                with request.urlopen(
                    "{}.zip".format(SHP_PATH)
                ) as response, open(ZIPFILE, "wb") as out_file:
                    shutil.copyfileobj(response, out_file)

            if not Path("{}{}.shp".format(DOWNLOAD_PATH, SHP_SLUG)).is_file():
                with zipfile.ZipFile(ZIPFILE, "r") as file:
                    file.extractall(DOWNLOAD_PATH)



# Imports from python.
import os


# Imports from Django.
from django.contrib.humanize.templatetags.humanize import ordinal


# Imports from other dependencies.
import geojson
import shapefile


class StateDistrictShapes(object):
    def get_state_district_shps(self, fips):
        SHP_SLUG = "cb_{}_us_cd{}_500k".format(self.YEAR, self.CONGRESS)
        DOWNLOAD_PATH = os.path.join(self.DOWNLOAD_DIRECTORY, SHP_SLUG)
        shape = shapefile.Reader(
            os.path.join(DOWNLOAD_PATH, "{}.shp".format(SHP_SLUG))
        )
        fields = shape.fields[1:]
        field_names = [f[0] for f in fields]
        district_records = [
            shp
            for shp in shape.shapeRecords()
            if dict(zip(field_names, shp.record))["STATEFP"] == fips
            # National district map
            or (
                fips == "00"
                and int(dict(zip(field_names, shp.record))["STATEFP"]) <= 56
            )
        ]
        features = []
        for shp in district_records:
            rec = dict(zip(field_names, shp.record))

            # Skip creating districts for territories and DC.
            if int(rec["STATEFP"]) > 56 or int(rec["STATEFP"]) == 11:
                continue

            code_key = "CD{}FP".format(self.CONGRESS)
            if int(rec[code_key]) == 0:
                label = "At-large congressional district".format(
                    ordinal(int(rec[code_key]))
                )
            else:
                label = "{} congressional district".format(
                    ordinal(int(rec[code_key]))
                )

            geometry = shp.shape.__geo_interface__
            geodata = {
                "type": "Feature",
                "geometry": geometry,
                "properties": {
                    "state": rec["STATEFP"],
                    "district": rec[code_key],
                    "name": label,
                },
            }
            features.append(geodata)
        threshold = self.THRESHOLDS["district"]
        return self.toposimplify(
            geojson.FeatureCollection(features), threshold
        )



# Imports from geography.
from geography.models import Geometry
from geography.serializers import GeometryListSerializer, GeometrySerializer
from geography.viewsets.base import BaseViewSet


class GeometryViewSet(BaseViewSet):
    queryset = Geometry.objects.all()

    def get_serializer_class(self):
        if self.action == "list":
            return GeometryListSerializer
        else:
            return GeometrySerializer



# Imports from geography.
from geography.models import Point
from geography.serializers import PointSerializer
from geography.viewsets.base import BaseViewSet


class PointViewSet(BaseViewSet):
    queryset = Point.objects.all()
    serializer_class = PointSerializer
